/*******************************************************
 Author:					Manfred Langemann
 mailto:					Manfred.Langemann t t-online.de
 Begin of project:			02.05.2020
 Latest version generated:	02.05.2020
 Processor:					ATmega32
 Filename:					QMC5883L.c
 Description:				Functions to access and read the Earth Magnetic Field Sensor IC QMC5883L
							The sensor is built on the GY-271 board (5V Power supply with I2C SCL and SDA lines)
 *******************************************************/
#include <stdio.h>
#include <avr/io.h>

#include "General.h"
#include "TWI_Master.h"
#include "QMC5883L.h"
/*******************************************************
 Public Function: QMC5883L_Init

 Purpose: Initiate the QMC5883L Earth Magnetic Field sensor.

 Input parameters: None
 
 Return value: uint8_t
	TRUE:	Initiation is correct
	FALSE:	Initiation failed
 *******************************************************/
uint8_t QMC5883L_Init (void)
	{
/*
** Initiate the TWI Master (I2C) interface with a rate of 100 000 symbols/s.
*/
	if (!TWIM_Init (100000)) return FALSE;
/*
** Soft Reset the QMC5883L Sensor.
*/
	if (!TWIM_Start (QMC5883L_ADDRESS, TWIM_WRITE))
		{
		TWIM_Stop ();
		return FALSE;
		}
	TWIM_Write (RESET_REGISTER);
	TWIM_Write (0x80);
	TWIM_Stop ();
/*
** Set/Reset Period of Sensor: Documentation states: "It is recommended that the register 0Bh is written by 0x01".
** When you do not set this bit, the sensor does not work and delivers random numbers.
*/	
	if (!TWIM_Start (QMC5883L_ADDRESS, TWIM_WRITE))
		{
		TWIM_Stop ();
		return FALSE;
		}
	TWIM_Write (PERIOD_REGISTER);
	TWIM_Write (0x01);
	TWIM_Stop ();
/*
** Set QMC5883L Sensor to:
**	- Mode				= Continuous
**	- Output Data Rate	= 100 Hz
**	- Full Scale		= 2 Gauss
**	- Over Sample Ratio = 512
*/
	if (!TWIM_Start (QMC5883L_ADDRESS, TWIM_WRITE))
		{
		TWIM_Stop ();
		return FALSE;
		}
	TWIM_Write (CONTROL_REGISTER);
	TWIM_Write (MODE_CONTINUOUS | MODE_OUTPUT_DATA_RATE_100HZ | MODE_RANGE_2G | MODE_OVER_SAMPLE_RATIO_512);
	TWIM_Stop ();
	
	return TRUE;
	}
/*******************************************************
 Public Function: QMC5883L_ReadMagneticData

 Purpose: Read from the QMC5883L sensor the 3-component Earth Magnetic Field.

 Input / Output parameters: 
	int16_t	*Data	Pointer to array of 3 x int16_t values
					Data[0] = X - Component
					Data[1] = Y - Component
					Data[2] = Z - Component

 Input parameters: 
	uint8_t			iContinueCheck
					TRUE: Continue to check data until it is ready
					FALSE: If data is not ready, return immediately with EMF_DATA_NOT_READY 
 
 Return value: uint8_t
	QMC5883L_OK					Read operation is correct
	QMC5883L_NO_ADDRESS			Cannot address I2C QMC5883L device
	QMC5883L_DATA_NOT_READY		Data not ready
	QMC5883L_TIMEOUT			Timeout is reached, when iContinueCheck is set to TRUE
 *******************************************************/
uint8_t QMC5883L_ReadMagneticData (int16_t *Data, uint8_t iContinueCheck)
	{
	uint8_t		iTimeout = 100;
	uint8_t		iStatus;
	uint16_t	iL, iM;
/*
** Check, whether data is ready by checking bit 0 of status register.
** 1. Send register address to read.
*/
Loop:
	if (!TWIM_Start (QMC5883L_ADDRESS, TWIM_WRITE))
		{
		TWIM_Stop ();
		return QMC5883L_NO_ADDRESS;
		}
	TWIM_Write (STATUS_REGISTER);
/*
** 2. Read the status register.
*/
	if (!TWIM_Start (QMC5883L_ADDRESS, TWIM_READ))
		{
		TWIM_Stop ();
		return QMC5883L_NO_ADDRESS;
		}
	iStatus = TWIM_ReadNack ();		// indicate by NACK, that transmission is completed
	TWIM_Stop ();
/*
** When bit 0 of status register is 1, data is ready to read, else not.
** When data is not ready, check timeout of number of try loops.
** If timeout is reached, return with EMF_TIMEOUT code.
*/
	if ((iStatus & 0x01) == 0)
		{
		if (iContinueCheck)
			{
			iTimeout--;
			if (iTimeout == 0) return QMC5883L_TIMEOUT;
			goto Loop;
			}
		return QMC5883L_DATA_NOT_READY;
		}
/*
** Read from the Sensor the Earh magnetic field data.
** 1. write to sensor from which register we want to start to read, which is register 0x00 = DATA_OUTPUT_REGISTER_START
*/
	if (!TWIM_Start (QMC5883L_ADDRESS, TWIM_WRITE))
		{
		TWIM_Stop ();
		return QMC5883L_NO_ADDRESS;
		}
	TWIM_Write (DATA_OUTPUT_REGISTER_START);
/*
** 2. read the content of the 6 data output registers (2xY, 2xY, 2xZ)
** By each read cycle, the sensor internal read register pointer will automatically be increased.
*/
	if (!TWIM_Start (QMC5883L_ADDRESS, TWIM_READ))
		{
		TWIM_Stop ();
		return QMC5883L_NO_ADDRESS;
		}

	iL = (uint16_t)TWIM_ReadAck ();		// LSB of X Component
	iM = (uint16_t)TWIM_ReadAck ();		// MSB of X Component
	*Data = (int16_t)(iM<<8 | iL);
	Data++;

	iL = (uint16_t)TWIM_ReadAck ();		// LSB of Y Component
	iM = (uint16_t)TWIM_ReadAck ();		// MSB of Y Component
	*Data = (int16_t)(iM<<8 | iL);
	Data++;	

	iL = (uint16_t)TWIM_ReadAck ();		// LSB of Z Component
	iM = (uint16_t)TWIM_ReadNack ();	// MSB of Z Component, indicate by NACK that transmission is completed
	*Data = (int16_t)(iM<<8 | iL);

	TWIM_Stop ();

	return QMC5883L_OK;
	}
/*******************************************************
 Public Function: Aileron_DoMeasurement

 Purpose: Report the error out of function QMC5883L_ReadMagneticData()

 Input parameters:
	uint8_t		iError

 Return value: void
 *******************************************************/
void QMC5883L_ReportError (uint8_t iError)
	{
	if (iError == QMC5883L_NO_ADDRESS)
		{
		printf ("Error: Cannot address QMC5883L !\n");
		return;
		}
	if (iError == QMC5883L_DATA_NOT_READY)
		{
		printf ("Error: QMC5883L data not ready !\n");
		return;
		}
	if (iError == QMC5883L_TIMEOUT)
		{
		printf ("Error: QMC5883L Timeout !\n");
		return;
		}
		
	return;
	}
