/*******************************************************
 Author:					Manfred Langemann
 mailto:					Manfred.Langemann t t-online.de
 Begin of project:			02.05.2020
 Latest version generated:	02.05.2020
 Processor:					ATmega32
 Filename:					QMC5883L.h
 Description:				Include file for QMC5883L.c
 *******************************************************/
#ifndef _QMC5883L
#define _QMC5883L
/*
** This file defines the data for addressing the QMC5883L Earth Magnetic Field Sensor, which is built on the GY-271 board (5V Power supply)
** The GY-271 board has four I/F lines and are connected to the ATmega32 as follows:
**
**	- +5V	--> Pin 10
**	- GND	--> Pin 11
**	- SDA	--> to TWI PORT PC1 (SDA), Pin 23
**	- SCL	--> to TWI PORT PC0 (SCL), Pin 22
**
** Pull up resistors for the two TWI/I2C lines are not necessary, because these are already on the GY-271 board with 4.7 KOhm.
**
** The I2C Address of the QMC5883L device is 0x0D.
*/
#define QMC5883L_ADDRESS				0x0D
/*
** These are the possible QMC5883L modes, in which the sensor can run.
** Combine them as necessary by bitwise OR = '|', e.g.: MODE_CONTINUOUS | MODE_OUTPUT_DATA_RATE_100HZ | MODE_RANGE_2G | MODE_OVER_SAMPLE_RATIO_512
*/
#define MODE_STANDBY					0x00
#define MODE_CONTINUOUS					0x01

#define MODE_OUTPUT_DATA_RATE_10HZ		0x00<<2
#define MODE_OUTPUT_DATA_RATE_50HZ		0x01<<2
#define MODE_OUTPUT_DATA_RATE_100HZ		0x02<<2
#define MODE_OUTPUT_DATA_RATE_200HZ		0x03<<2

#define MODE_RANGE_2G					0x00<<4
#define MODE_RANGE_8G					0x01<<4

#define MODE_OVER_SAMPLE_RATIO_512		0x00<<6
#define MODE_OVER_SAMPLE_RATIO_256		0x01<<6
#define MODE_OVER_SAMPLE_RATIO_128		0x02<<6
#define MODE_OVER_SAMPLE_RATIO_64		0x03<<6
/*
** These are the QMC5883L registers, which can be addressed.
*/
#define DATA_OUTPUT_REGISTER_START		0x00	// Read only

#define DATA_OUTPUT_REGISTER_X_LSB		0x00	// Read only
#define DATA_OUTPUT_REGISTER_X_MSB		0x01	// Read only
#define DATA_OUTPUT_REGISTER_Y_LSB		0x02	// Read only
#define DATA_OUTPUT_REGISTER_Y_MSB		0x03	// Read only
#define DATA_OUTPUT_REGISTER_Z_LSB		0x04	// Read only
#define DATA_OUTPUT_REGISTER_Z_MSB		0x05	// Read only


#define STATUS_REGISTER					0x06	// Read only
#define TEMPERATURE_REGISTER_LSB		0x07	// Read only
#define TEMPERATURE_REGISTER_MSB		0x08	// Read only
#define CONTROL_REGISTER				0x09	// Read / Write
#define RESET_REGISTER					0x0A	// Read / Write
#define PERIOD_REGISTER					0x0B	// Read / Write

enum {
	QMC5883L_OK,				// Read operation is correct
	QMC5883L_NO_ADDRESS,		// Cannot address I2C devide
	QMC5883L_DATA_NOT_READY,	// Data not ready
	QMC5883L_TIMEOUT,		// Timeout
	};
/*
** Public Functions.
*/
uint8_t QMC5883L_Init (void);
uint8_t QMC5883L_ReadMagneticData (int16_t *Data, uint8_t iContinueCheck);
void QMC5883L_ReportError (uint8_t iError);

#endif
