/*******************************************************
 Author:					Manfred Langemann
 mailto:					Manfred.Langemann t t-online.de
 Begin of project:			04.01.2008
 Latest version generated:	12.02.2020 --> see function TWIM_Init()
 Filename:					TWI_Master.c
 Description:				TWI Master functions
 --------------------------------------------------------
 Program Example
 --------------------------------------------------------

int main (void)
	{
	uint8_t		i=0;
	uint8_t		j=0;
	uint8_t		Data[8];
	uint8_t		SlaveAddress = 15;

// Clear any interrupt
	cli ();

// Wait 1 second for POR
	Delay_ms (1000);

// Initiate RS232
	RS232_Init ();
	printf ("Hello world...\n");

// Initiate TWI Master with bitrate of 100000 Hz
	TWIM_Init (100000);

// Endless loop
	while (1)
		{
// Read byte(s) from the slave.
// It is implicite assumed, that the slave will send
// 8 bytes.
    	if (!TWIM_Start (SlaveAddress, TWIM_READ))
			{
			TWIM_Stop ();
			printf ("Could not start TWI Bus for READ\n");
	    	}
		else
			{
			for (i=0;i<7;i++)
				{
	        	Data[i] = TWIM_ReadAck ();
				}
        	Data[7] = TWIM_ReadNack ();
			printf ("Reading Byte %d: %d\n", i, Data[i]);
			TWIM_Stop ();
			Delay_ms (1000);
			}

// Write byte(s) to the slave.
// It is implicite assumed, that the slave will
// accepts 8 bytes

    	if (!TWIM_Start (SlaveAddress, TWIM_WRITE))
			{
			TWIM_Stop ();
			printf ("Could not start TWI Bus for WRITE\n");
	    	}
		else
			{
			for (i=0;i<8;i++)
				{
	        	TWIM_Write (j++);
				printf ("Byte %d sent\n", j);
				}
			TWIM_Stop ();
			Delay_ms (1000);
			}

// Do something else
		i++;
		}

	return 0;
	}
 ********************************************************/
#include <stdio.h>
#include <avr/interrupt.h>

#include "General.h"
#include "TWI_Master.h"

/****************************************************************************
  Private TWI Master functions
****************************************************************************/
uint8_t TWIM_Start (uint8_t address, uint8_t TWIM_Type);
void	TWIM_Stop (void);
uint8_t	TWIM_Write (uint8_t byte);
uint8_t	TWIM_ReadAck (void);
uint8_t	TWIM_ReadNack (void);
uint8_t	TWIM_Read (uint8_t Ack);

/****************************************************************************
  TWI State codes
****************************************************************************/
// General TWI Master status codes                      
#define TWI_START					0x08  // START has been transmitted  
#define TWI_REP_START				0x10  // Repeated START has been transmitted
#define TWI_ARB_LOST				0x38  // Arbitration lost

// TWI Master Transmitter status codes                      
#define TWI_MTX_ADR_ACK				0x18  // SLA+W has been tramsmitted and ACK received
#define TWI_MTX_ADR_NACK			0x20  // SLA+W has been tramsmitted and NACK received 
#define TWI_MTX_DATA_ACK			0x28  // Data byte has been tramsmitted and ACK received
#define TWI_MTX_DATA_NACK			0x30  // Data byte has been tramsmitted and NACK received 

// TWI Master Receiver status codes  
#define TWI_MRX_ADR_ACK				0x40  // SLA+R has been tramsmitted and ACK received
#define TWI_MRX_ADR_NACK			0x48  // SLA+R has been tramsmitted and NACK received
#define TWI_MRX_DATA_ACK			0x50  // Data byte has been received and ACK tramsmitted
#define TWI_MRX_DATA_NACK			0x58  // Data byte has been received and NACK tramsmitted

// TWI Slave Transmitter status codes
#define TWI_STX_ADR_ACK				0xA8  // Own SLA+R has been received; ACK has been returned
#define TWI_STX_ADR_ACK_M_ARB_LOST	0xB0  // Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
#define TWI_STX_DATA_ACK			0xB8  // Data byte in TWDR has been transmitted; ACK has been received
#define TWI_STX_DATA_NACK			0xC0  // Data byte in TWDR has been transmitted; NOT ACK has been received
#define TWI_STX_DATA_ACK_LAST_BYTE	0xC8  // Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received

// TWI Slave Receiver status codes
#define TWI_SRX_ADR_ACK				0x60  // Own SLA+W has been received ACK has been returned
#define TWI_SRX_ADR_ACK_M_ARB_LOST	0x68  // Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
#define TWI_SRX_GEN_ACK				0x70  // General call address has been received; ACK has been returned
#define TWI_SRX_GEN_ACK_M_ARB_LOST	0x78  // Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
#define TWI_SRX_ADR_DATA_ACK		0x80  // Previously addressed with own SLA+W; data has been received; ACK has been returned
#define TWI_SRX_ADR_DATA_NACK		0x88  // Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
#define TWI_SRX_GEN_DATA_ACK		0x90  // Previously addressed with general call; data has been received; ACK has been returned
#define TWI_SRX_GEN_DATA_NACK		0x98  // Previously addressed with general call; data has been received; NOT ACK has been returned
#define TWI_SRX_STOP_RESTART		0xA0  // A STOP condition or repeated START condition has been received while still addressed as Slave

// TWI Miscellaneous status codes
#define TWI_NO_STATE				0xF8  // No relevant state information available; TWINT = 0
#define TWI_BUS_ERROR				0x00  // Bus error due to an illegal START or STOP condition

/*******************************************************
 Public Function: TWIM_Init

 Purpose: Initialise the TWI Master Interface

 Input Parameter:
 	- uint32_t	TWI_Bitrate (Hz)

 Return Value: uint8_t
 	- FALSE:	Bitrate too high
 	- TRUE:		Bitrate OK
	 
Bekannte Fehler:

	siehe: https://www.mikrocontroller.net/articles/AVR_TWI

	Bei einigen ATMEGA Varianten (z.B. ATMEGA8) hat die TWI Implementierung einen Bug.
	Nach dem Aufwachen aus einem Sleep Modus funktioniert der TWI nicht mehr. Erst ein Reset mit:

		TWCR &= ~((1 << TWSTO) | (1 << TWEN));
		TWCR |= (1 << TWEN);

	nach dem Aufwachen lst das Problem.

	Nach dem Senden eines "Stop" mu vor erneutem "Start" das TWSTO Bit im TWCR Register auf "0" abgefragt werden.
	Wenn dies noch eins enthlt ist der Stop nicht am Bus wirksam.
	Ein erneuter Start innerhalb dieser Situation bringt statt der "Start" Rckmeldung "$08" die "Restart" Rckmeldung "$10".

*******************************************************/
uint8_t TWIM_Init (uint32_t TWI_Bitrate)
	{
	//uint16_t	iTWBR;
/*
** Set TWI Master bitrate.
** Note: The bitrate of the TWI is not the data rate of the TWI interface.
** The data rate is the bit rate divided by 9 and multiplied by 8, because for each byte the TWI will need 9 TWI clock pulses.
**
** Slave operation does not depend on Bit Rate or Prescaler settings,
** but the CPU clock frequency in the Slave must be at least 16 times higher than the SCL_frequency.
**
** It is assumed here, that the slave has the same CPU frequency (CPU_clock) than the master or at least 10 MHz.
** If input parameter TWI_Bitrate is too high, then decrease the bitrate until it matches with 16 MHz.
** The TWI Bit Rate Prescaler value shall be TWPS=1, i.e. setting the TWPS bits to: TWPS0=0 and TWPS1=0.
**
** Using a 16 MHz crystal (CPU_clock) in the slave allows to set the TWI_Bitrate up to a value of 1 MHz.
** Using a 10 MHz crystal (CPU_clock) in the slave allows to set the TWI_Bitrate up to a value of 0,625 MHz.
**
** Now Check the required TWI Bitrate.
*/
	if (F_CPU/16 < TWI_Bitrate) return FALSE;
/*
** Set TWSP0 and TWSP1 to 0 --> Prescaler = 1.
*/
	TWSR = 0;
/*
** The formula to set the TWI Bit Rate register TWBR is as follows:
** SCL_frequ = CPU_clock / (16 + (2*TWBR * 4^TWPS)), where TWPS (TWPS1 and TWPS0) is the value of the prescaler bits in the TWI Status Register, i.e.
** - 00 (TWPS=0) --> Prescaler is 1
** - 01 (TWPS=1) --> Prescaler is 4
** - 10 (TWPS=2) --> Prescaler is 16
** - 11 (TWPS=3) --> Prescaler is 64
** (see formula in sub-chapter "Bit Rate Generator Unit" in chapter "Two-Wire Serial Interface" in Atmel Documentation is)
** Because we set the TWI Bit Rate Prescaler to 1, it follows with 4^0 = 1:
** SCL_frequ = CPU_clock / (16 + (2*TWBR * 4^0)) = CPU_clock / (16 + (2*TWBR))
**
** Reorganising the formula to get the required value for the TWBR register is then:
**	TWBR = ((CPU_clock/SCL_frequ)-16)/2
** Assuming a CPU_clock = 16 MHz and a TWI Bitrate of 300.000 bit/s we get a TWBR value of 18.
** Assuming a CPU_clock = 16 MHz and a TWI Bitrate of 100.000 bit/s we get a TWBR value of 72.
**
** Note: TWBR values of less than 10 shall be avoided !
**
** Es gab Diskussionen auf:
** https://www.mikrocontroller.net/topic/87597#new
** https://rn-wissen.de/wiki/index.php/TWI
**
** Note:
**	- F_CPU = CPU_clock
**	- TWI_Bitrate = SCL_frequ
*/
	TWBR = ((F_CPU/TWI_Bitrate)-16)/2; 
	//iTWBR = TWBR;
	//printf ("TWBR = %i\n", iTWBR);
	return TRUE;
	}
/*******************************************************
 Public Function: TWIM_Start

 Purpose: Start the TWI Master Interface

 Input Parameter:
 	- uint8_t	Device address
 	- uint8_t	Type of required Operation:
				TWIM_READ: Read data from the slave
				TWIM_WRITE: Write data to the slave

 Return Value: uint8_t
  	- TRUE:		OK, TWI Master accessible
 	- FALSE:	Error in starting TWI Master
*******************************************************/
uint8_t TWIM_Start (uint8_t Address, uint8_t TWIM_Type)
	{
	uint8_t		twst;
	//printf ("Entering TWIM_Start\n");
/*
** Send START condition.
*/
	TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);
/*
** Wait until transmission completed.
*/
	while (!(TWCR & (1<<TWINT)));
	//printf ("Start Completed OK\n");
/*
** Check value of TWI Status Register. Mask prescaler bits.
*/
	twst = TWSR & 0xF8;
	if ((twst != TWI_START) && (twst != TWI_REP_START)) return FALSE;
	//printf ("Start Status OK\n");
/*
** Send device address.
*/
	TWDR = (Address<<1) + TWIM_Type;
	TWCR = (1<<TWINT) | (1<<TWEN);
/*
** Wait until transmission completed and ACK has been received.
*/
	while (!(TWCR & (1<<TWINT)));
	//printf ("Send Address Completed\n");
/*
** Check value of TWI Status Register. Mask prescaler bits.
** Value must be ACK in receive or transmit mode (MTX or MRX)
*/
	twst = TWSR & 0xF8;
	//printf ("TWSR=%xh\n", twst);
	if ((twst != TWI_MTX_ADR_ACK) && (twst != TWI_MRX_ADR_ACK)) return FALSE;
	//printf ("ACK or NACK received OK\n");

	return TRUE;
	}
/*******************************************************
 Public Function: TWIM_Stop

 Purpose: Stop the TWI Master

 Input Parameter: None

 Return Value: None
*******************************************************/
void TWIM_Stop (void)
	{
/*
** Send stop condition.
*/
	TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
/*
** Wait until stop condition is executed and bus released.
*/
	while (TWCR & (1<<TWINT));
	}
/*******************************************************
 Public Function: TWIM_Write

 Purpose: Write a byte to the slave

 Input Parameter:
 	- uint8_t	Byte to be sent

 Return Value: uint8_t
  	- TRUE:		OK, Byte sent
 	- FALSE:	Error in byte transmission
*******************************************************/
uint8_t TWIM_Write (uint8_t byte)
	{
	uint8_t   twst;
/*
** Send data to the previously addressed device
*/
	TWDR = byte;
	TWCR = (1<<TWINT) | (1<<TWEN);
/*
** Wait until transmission completed.
*/
	while (!(TWCR & (1<<TWINT)));
/*
** Check value of TWI Status Register. Mask prescaler bits
** Allowed values are only TWI_MTX_DATA_ACK or TWI_MTX_DATA_NACK
*/
	twst = TWSR & 0xF8;
	if ((twst != TWI_MTX_DATA_ACK) && (twst != TWI_MTX_DATA_NACK)) return FALSE;

	return TRUE;
	}
/*******************************************************
 Public Function: TWIM_ReadAck

 Purpose: Read a byte from the slave and request next byte

 Input Parameter: None

 Return Value: uint8_t
	- uint8_t	Read byte
*******************************************************/
uint8_t TWIM_ReadAck (void)
	{
	TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA);
	while (!(TWCR & (1<<TWINT)));

	return TWDR;
	}
/*******************************************************
 Public Function: TWIM_ReadNack

 Purpose: Read the last byte from the slave

 Input Parameter: None

 Return Value: uint8_t
  	- uint8_t	Read byte
*******************************************************/
uint8_t TWIM_ReadNack (void)
	{
	TWCR = (1<<TWINT) | (1<<TWEN);
	while (!(TWCR & (1<<TWINT)));
	
	return TWDR;
	}
/*******************************************************
 Public Function: TWIM_Read

 Purpose: Read a byte from the slave

 Input Parameter:
  	- uint8_t	Ack
		Ack == TRUE  --> send ACK  --> Slave expect to send more bytes
		Ack == FALSE --> send NACK --> Slave sends last byte

 Return Value: uint8_t
		- uint8_t	Read byte
*******************************************************/
uint8_t	TWIM_Read (uint8_t Ack)
	{
	uint8_t		i;

	i = (1<<TWINT) | (1<<TWEN);
	if (Ack)
		{
		i = i | (1<<TWEA);
		}
	TWCR = i;
	while (!(TWCR & (1<<TWINT)));

	return TWDR;
	}
///*******************************************************
 //Public Function: TWIM_ReadDataFromTWI
//
 //Purpose: Read command from the TWI Interface and place data
 			//to CommandStruct.
			 //
			//Note: First received byte indicates the number of
			//following data bytes to be read from TWI interface.
//
 //Input Parameter:
 	//- global FRAME_STRUCT	 TXRX_Frame	Where to write the received data
 //
 //Return Value: uint8_t
  	//TRUE:	Command read correctly
  	//FALSE:	Error in reading command
//
//*******************************************************/
//uint8_t TWIM_ReadDataFromTWI ()
	//{
	//uint8_t		k, i;
	//uint8_t		SlaveAddress;
	//uint8_t		NumBytes;
	//uint8_t		Ack = TRUE;
///*
//** Extract the Slave Address (Controller ID number) from the Frame Header Packet.
//*/
	//SlaveAddress = FrameHeader_Get_Controller ();
	////SlaveAddress = FRAMEHEADER_GET_CONTROLLER(TXRX_Frame.FrameHeaderPacket);
///*
//** Start to read from TWI with Slave Address.
//*/
	//if (!TWIM_Start (SlaveAddress, TWIM_READ))
		//{
		//TWIM_Stop ();
		//return FALSE;
		//}
	////printf ("Reading Start OK\n");
///*
//** Read from TWI, how many bytes to receive.
//** NumBytes is 5 Bytes of the Frame Header Packet plus the optional Data Bytes.
//*/
	//NumBytes = TWIM_Read (Ack);
	////printf ("Reading NumBytes OK\n");
///*
//** Clear the content of the Frame Header Packet, because we only read 5 Bytes of the 8 Byte Size value.
//*/
	//TXRX_Frame.FrameHeaderPacket = 0;
///*
//** First read the 5 bytes of the Frame Header Packet.
//** Note: if NumBytes is only 5, and having received 4 Bytes, Ack will be set to FALSE,
//** which means we are reading the last byte.
//*/
	//for (i=0; i<FRAME_HEADER_PACKET_SIZE; i++)
		//{
		//if (i == (NumBytes-1)) Ack = FALSE;
		//k = TWIM_Read (Ack);
		//TXRX_FrameHeader_Buf[i] = k;
		//}
	////printf ("Reading 5 Bytes OK\n");
///*
//** If we also have to read optional data bytes, then do it here.
//** Note: the if loop only starts, if NumBytes > 0.
//*/
	//NumBytes = NumBytes - FRAME_HEADER_PACKET_SIZE;
	//for (i=0; i<NumBytes; i++)
		//{
		//if (i == (NumBytes-1)) Ack = FALSE;
		//k = TWIM_Read (Ack);
		//TXRX_Frame.FrameDataPacket[i] = k;
		//}
	////printf ("Reading DataBytes OK\n");
///*
//** Stop TWI communication.
//*/	
	//TWIM_Stop ();
//
	//return TRUE;
	//}
///*******************************************************
 //Public Function: TWIM_WriteDataToTWI
//
 //Purpose: Write command to the TWI Interface
 //
 //Input Parameter:
 	//- global FRAME_STRUCT	 TXRX_Frame	 Content of command to write
 //
 //Return Value: uint8_t
   	//TRUE:	Command sent correctly
   	//FALSE:	Error in sending command
//
//*******************************************************/
//uint8_t TWIM_WriteDataToTWI (void)
	//{
	//uint8_t		SlaveAddress;
	//uint8_t		NumBytes;
	//uint8_t		i;
	//uint8_t		value;
///*
//** Extract the Slave Address (Controller ID number) from the Frame Header Packet.
//*/	
	//SlaveAddress = FrameHeader_Get_Controller ();
	////SlaveAddress = FRAMEHEADER_GET_CONTROLLER(TXRX_Frame.FrameHeaderPacket);
///*
//** Start TWI for WRITE.
//*/
   	//if (!TWIM_Start (SlaveAddress, TWIM_WRITE)) goto Error;
	////printf ("TWIM_Start was OK\n");
///*
//** Write to the slave the number of total bytes we will transmit, excluding this byte.
//*/
	//NumBytes = FRAME_HEADER_PACKET_SIZE + FrameHeader_Get_Num_Data_Bytes ();
	////NumBytes = FRAME_HEADER_PACKET_SIZE + FRAMEHEADER_GET_NUM_DATA_BYTES(TXRX_Frame.FrameHeaderPacket);
	//if (!TWIM_Write (NumBytes)) goto Error;
	////printf ("NumBytes OK\n");
///*
//** Write to the slave 5 Bytes, representing the Frame Header Packet.
//*/
	//for (i=0; i<FRAME_HEADER_PACKET_SIZE; i++)
		//{
		//value = TXRX_FrameHeader_Buf[i];
		//if (!TWIM_Write (value)) goto Error;
		//}
	////printf ("5 Bytes OK\n");
///*
//** If we have to write also Data Bytes, then do this here.
//** Note: the "if" loop only starts, if NumBytes > 0.
//*/
	//NumBytes = NumBytes - FRAME_HEADER_PACKET_SIZE;
	//for (i=0; i<NumBytes; i++)
		//{
		//value = TXRX_Frame.FrameDataPacket[i];
		//if (!TWIM_Write (value)) goto Error;
		//}
	////if (NumBytes >0) printf ("Data Bytes OK\n");
///*
//** Stop transmission and return.
//*/		
	//TWIM_Stop ();
	//return TRUE;
///*
//** Error return.
//*/
//Error:
	//TWIM_Stop ();
	//return FALSE;
	//}
